﻿using System;
using System.Threading;

/* Raima.Rdm is the system module that contains the ADO.NET
 * data provider for RDM */
using Raima.Rdm;

/* EXAMPLE - SQL11 for C# using ADO.NET */
/* This example shows multi-user access to a database using separate threads.
 * Two writer threads are created.  The first begins inserting values into
 * the table.  The second waits until the table contains more than 100 rows
 * and then begins deleting the oldest.  A reader thread is also created
 * that prints out the current oldest and newest entries every second.  Each
 * thread uses regular locking.
 * 
 * The following methods are illustrated:
 *     Raima.Rdm.RdmConnection:
 *         ChangeDatabase(string) Method
 *         Close() Method
 *         CreateCommand() Method
 *         Open() Method
 *         RdmConnection() Constructor
 *     Raima.Rdm.RdmCommand:
 *         CommandText Property
 *         ExecuteNonQuery() Method
 *         ExecuteReader() Method
 *         Parameters Property
 *     Raima.Rdm.RdmDataReader:
 *         Item[int] Property
 *         Read() Method
 *     Raima.Rdm.RdmParameterCollection
 *         Add(string, RdmType)
 *         Item[string] Property
 * 
 * They are fully document in http://docs.raima.com/rdm/.
 * 
 * For simplicity this example does not catch any thrown execptions
 * but good programming practices would dictate that this occur in
 * a full application.
 */

namespace SQL11
{
    class Program
    {
        static void DropDatabase(
            RdmConnection conn,
            string name)
        {
            RdmCommand cmd = conn.CreateCommand();

            /* Attempt to drop this database.  If the database does not
             * already exist this will fail and we simply ignore the
             * failure for this sample. */
            try
            {
                cmd.CommandText = "DROP DATABASE " + name;
                cmd.ExecuteNonQuery();

                cmd.CommandText = "COMMIT";
                cmd.ExecuteNonQuery();
            }
            catch (Exception)
            {
            }
        }

        static void CreateThisDatabase(
            RdmConnection conn,
            string        name,
            string[]      cmds)
        {
            RdmCommand cmd = conn.CreateCommand();

            /* Create the specified database.  Make sure the database is
             * closed when we are done so other databases can be created
             * after this if necessary. */
            cmd.CommandText = "CREATE DATABASE " + name;
            cmd.ExecuteNonQuery();

            foreach (string c in cmds)
            {
                cmd.CommandText = c;
                cmd.ExecuteNonQuery();
            }

            cmd.CommandText = "COMMIT";
            cmd.ExecuteNonQuery();

            cmd.CommandText = "CLOSE " + name;
            cmd.ExecuteNonQuery();
        }

        static void CreateDatabases(
            RdmConnection conn)
        {
            string[] cmds =
            {
                "CREATE TABLE Table1 (id INTEGER PRIMARY KEY)"
            };

            /* Drop any existing copy of the database */
            DropDatabase(conn, "SQL11");

            /* Attempt to create the database */
            CreateThisDatabase(conn, "SQL11", cmds);
        }

        static bool shutdown = false;

        static void InserterThreadProc()
        {
            /* Create a connection for this thread */
            RdmConnection conn = new RdmConnection("host=local;database=SQL11;openmode=s");
            conn.Open();

            /* Create a command to do the insert */
            RdmCommand cmd = conn.CreateCommand();
            cmd.CommandText = "INSERT INTO Table1 VALUES :id";
            cmd.Parameters.Add("id", RdmType.Int32);

            /* Loop until a shutdown is requested */
            for (int ii = 0; !shutdown; ii++)
            {
                /* Insert the next value */
                cmd.Parameters["id"].Value = ii;
                cmd.ExecuteNonQuery();
            }

            /* When a shutdown is requested, close the connection and exit */
            conn.Close();
        }

        static void DeleterThreadProc()
        {
            /* Create a connection for this thread */
            RdmConnection conn = new RdmConnection("host=local;database=SQL11;openmode=s");
            conn.Open();

            /* Create a command to check the number of rows, one to find the
             * 'smallest' row, and one to delete */
            RdmCommand chkCmd = conn.CreateCommand();
            RdmCommand minCmd = conn.CreateCommand();
            RdmCommand delCmd = conn.CreateCommand();

            chkCmd.CommandText = "SELECT COUNT(id) FROM Table1";
            minCmd.CommandText = "SELECT MIN(id) FROM Table1";
            delCmd.CommandText = "DELETE FROM Table1 WHERE id = :minid";
            RdmParameter parm = delCmd.Parameters.Add("minid", RdmType.Int32);

            /* Loop until a shutdown is requested */
            while (!shutdown)
            {
                /* Query the number of rows in the table */
                RdmDataReader countRdr = chkCmd.ExecuteReader();
                if (countRdr.Read())
                {
                    long count = (long) countRdr[0];

                    /* There should be only one row, so this closes the reader */
                    countRdr.Read();

                    if (count > 100)
                    {
                        /* If there at least 100 rows, get the smallest and delete it */
                        RdmDataReader minRdr = minCmd.ExecuteReader();
                        if (minRdr.Read())
                        {
                            parm.Value = (int) minRdr[0];
                            delCmd.ExecuteNonQuery();

                            /* There should be only one row, so this closes the reader */
                            minRdr.Read();
                        }
                    }
                }
            }

            conn.Close();
        }

        static void ReaderThreadProc()
        {
            /* Create a connection for this thread */
            RdmConnection conn = new RdmConnection("host=local;database=SQL11;openmode=s");
            conn.Open();

            /* Create a command to do the read */
            RdmCommand cmd = conn.CreateCommand();
            cmd.CommandText = "SELECT MIN(id), MAX(id) FROM Table1";

            /* Loop until a shutdown is requested */
            while (!shutdown)
            {
                /* Wait one second */
                Thread.Sleep(1000);

                /* Query the mininum and maximum values */
                RdmDataReader rdr = cmd.ExecuteReader();
                if (rdr.Read())
                {
                    Console.WriteLine("Current minimum = " + rdr[0] + " and current maximum = " + rdr[1]);
                    rdr.Read();
                }
            }

            conn.Close();
        }

        static void Main(string[] args)
        {
            /* Create the connection with a simple connection string */
            RdmConnection conn = new RdmConnection("host=local;openmode=s");
            conn.Open();

            CreateDatabases(conn);

            /* Create three threads one to insert, one to delete, and one to 
             * show the current contents */
            Thread inserter = new Thread(new ThreadStart(InserterThreadProc));
            Thread deleter = new Thread(new ThreadStart(DeleterThreadProc));
            Thread reader = new Thread(new ThreadStart(ReaderThreadProc));

            Console.WriteLine("Starting threads");
            Console.WriteLine("   Press <Enter> when you are ready to terminate them");

            /* Start each of the threads */
            inserter.Start();
            deleter.Start();
            reader.Start();

            /* Wait for user to hit <Enter> */
            Console.ReadLine();

            /* Indicate to the threads that they should shut down */
            shutdown = true;

            /* Wait for each thread to finish */
            inserter.Join();
            deleter.Join();
            reader.Join();

            Console.WriteLine("All threads have terminated");

            conn.Close();

            Console.WriteLine("Hit <Enter> to continue");
            Console.ReadLine();
        }
    }
}
